import tensorflow as tf 
import numpy as np

def calculate_accuracy_classification(input_tensor, label, threshold=0.5):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [description]
        label {[type]} -- [description]
    
    Keyword Arguments:
        threshold {float} -- [description] (default: {0.5})
    """
    matches = tf.equal(tf.argmax(input_tensor, 1), tf.argmax(label, 1))
    accuracy = tf.reduce_mean(tf.cast(matches, tf.float32))
    return accuracy

def calculate_accuracy_mask(input_tensor, label, threshold=0.5):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        threshold {float} -- [description] (default: {0.5})
    
    Returns:
        [Scalar] -- [Accuracy between output graph and real target]
    """
    mask = tf.fill(tf.shape(label), 1.)
    input_tensor = tf.reshape(input_tensor, [tf.shape(input_tensor)[0], -1])
    label = tf.reshape(label, [tf.shape(label)[0], -1])

    input_tensor = tf.math.greater(input_tensor, tf.convert_to_tensor(np.array(threshold), tf.float32))
    input_tensor = tf.cast(input_tensor, tf.float32)
    label = tf.math.greater(label, tf.convert_to_tensor(np.array(threshold), tf.float32))
    label = tf.cast(label, tf.float32)

    error = tf.reduce_sum(tf.abs(input_tensor - label)) / (tf.reduce_sum(mask) + 0.0001)
    acc = 1. - error
    return acc

def calculate_loss(input_tensor, label, type_loss_function='sigmoid_crossentropy_mean', penalize_class=0, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [description]
        label {[type]} -- [description]
    
    Keyword Arguments:
        penalize_class {int} -- [description] (default: {0})
    """
    if names!=None:
        names = str(names)
    else:
        names= 'loss'

    if type_loss_function == 'mse_loss_mean':
        loss = tf.compat.v1.losses.mean_squared_error(labels=label, predictions=input_tensor)
        loss = tf.reduce_mean(loss)
    elif type_loss_function == 'mse_loss_sum':
        loss = tf.compat.v1.losses.mean_squared_error(labels=label, predictions=input_tensor)
        loss = tf.reduce_sum(loss)
    elif type_loss_function == 'softmax_crosentropy_mean':
        loss = tf.compat.v2.nn.softmax_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
        loss = tf.reduce_mean(loss)
    elif type_loss_function == 'softmax_crosentropy_sum':
        loss = tf.compat.v2.nn.softmax_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
        loss = tf.reduce_sum(loss)
    elif type_loss_function == 'sigmoid_crossentropy_mean':
        loss = tf.compat.v2.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
        loss = tf.reduce_mean(loss)
    elif type_loss_function == 'sigmoid_crossentropy_sum':
        loss = tf.compat.v2.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
        loss = tf.reduce_sum(loss)
    else:
        loss = tf.compat.v2.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
        loss = tf.reduce_mean(loss)
    return loss

def mse_mean(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'mse_mean_loss'

    loss = tf.compat.v1.losses.mean_squared_error(labels=label, predictions=input_tensor)
    loss = tf.reduce_mean(loss)
    return loss


def mse_sum(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'mse_sum_loss'

    loss = tf.compat.v1.losses.mean_squared_error(labels=label, predictions=input_tensor)
    loss = tf.reduce_sum(loss)
    return loss


def softmax_crosentropy_mean(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'softmax_crossentropy_mean_loss'

    loss = tf.compat.v2.nn.softmax_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
    loss = tf.reduce_mean(loss)
    return loss


def softmax_crosentropy_sum(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'softmax_crossentropy_sum_loss'

    loss = tf.compat.v2.nn.softmax_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
    loss = tf.reduce_sum(loss)
    return loss


def sigmoid_crossentropy_mean(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'sigmoid_crossentropy_mean_loss'

    loss = tf.compat.v2.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
    loss = tf.reduce_mean(loss)
    return loss


def sigmoid_crossentropy_sum(input_tensor, label, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [A Tensor representing output graph]
        label {[type]} -- [A Tensor representing real target]
    
    Keyword Arguments:
        names {[type]} -- [Name for metrics function] (default: {None})
    
    Returns:
        [Scalar] -- [Loss of Value between output graph and real target]
    """
    if names!=None:
        names = str(names)
    else:
        names= 'sigmoid_crossentropy_sum_loss'

    loss = tf.compat.v2.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input_tensor, name=names)
    loss = tf.reduce_sum(loss)
    return loss


import json
import random
import tensorflow as tf
from poda.layers.convolutional import *
from poda.layers.dense import *
from poda.layers.activation import *
from poda.object_detector.utils import *


class YoloV3(ObjectDetector):
    def __init__(self,
                 dataset_folder_path,
                 label_folder_path,  
                 num_of_class,
                 input_height=416, 
                 input_width=416, 
                 grid_height1=32, 
                 grid_width1=32, 
                 grid_height2=16, 
                 grid_width2=16, 
                 grid_height3=8, 
                 grid_width3=8,
                 objectness_loss_alpha=2., 
                 noobjectness_loss_alpha=1., 
                 center_loss_alpha=1., 
                 size_loss_alpha=1., 
                 class_loss_alpha=1.,
                 add_modsig_toshape=False,
                 anchor = [(10, 13), (16, 30), (33, 23), (30, 61), (62, 45), (59, 119), (116, 90), (156, 198), (373, 326)],
                 dropout_rate = 0.8,
                 leaky_relu_alpha = 0.1):
        """[summary]
        
        Arguments:
            label_folder_path {[type]} -- [description]
            dataset_folder_path {[type]} -- [description]
        
        Keyword Arguments:
            input_height {int} -- [description] (default: {512})
            input_width {int} -- [description] (default: {512})
            grid_height {int} -- [description] (default: {128})
            grid_width {int} -- [description] (default: {128})
            output_depth {int} -- [description] (default: {5})
            objectness_loss_alpha {[type]} -- [description] (default: {1.})
            noobjectness_loss_alpha {[type]} -- [description] (default: {1.})
            center_loss_alpha {[type]} -- [description] (default: {0.})
            size_loss_alpha {[type]} -- [description] (default: {0.})
            class_loss_alpha {[type]} -- [description] (default: {0.})
        """

        super(Yolo, self).__init__(num_of_class=num_of_class,
                                        input_height=input_height, 
                                        input_width=input_width, 
                                        grid_height1=grid_height1, 
                                        grid_width1=grid_width1, 
                                        grid_height2=grid_height2, 
                                        grid_width2=grid_width2, 
                                        grid_height3=grid_height3, 
                                        grid_width3=grid_height3,
                                        objectness_loss_alpha=objectness_loss_alpha, 
                                        noobjectness_loss_alpha=noobjectness_loss_alpha, 
                                        center_loss_alpha=center_loss_alpha, 
                                        size_loss_alpha=size_loss_alpha, 
                                        class_loss_alpha=class_loss_alpha,
                                        add_modsig_toshape=add_modsig_toshape,
                                        anchor = anchor,
                                        dropout_rate = dropout_rate,
                                        leaky_relu_alpha = leaky_relu_alpha)

        self.label_folder_path = label_folder_path
        self.dataset_folder_path = dataset_folder_path
        self.dataset_file_list = get_filenames(self.dataset_folder_path)
        random.shuffle(self.dataset_file_list)
        
        print ("------------------------INFO-------------------")
        print ("Image Folder: " + self.dataset_folder_path)
        print ("Number of Image: " + str(len(self.dataset_file_list)))
        print ("-----------------------------------------------")

        self.all_label_target_np = None

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, 3))
        self.output_placeholder1 = tf.placeholder(tf.float32, shape=(None, 13, 13, 3*(5 + num_of_class)))
        self.output_placeholder2 = tf.placeholder(tf.float32, shape=(None, 26, 26, 3*(5 + num_of_class)))
        self.output_placeholder3 = tf.placeholder(tf.float32, shape=(None, 52, 52, 3*(5 + num_of_class)))

    
    def read_target(self, file_path):
        """[summary]
        
        Arguments:
            file_path {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        target = self.read_yolo_labels(file_path)
        return target


    def build_net(self, input_tensor, 
                  network_type='big', 
                  is_training=False):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            network_type {str} -- [description] (default: {'big'})
            is_training {bool} -- [description] (default: {False})
        """
        with tf.variable_scope('yolo_v3_model'):
            self.build_yolov3_net(inputs=input_tensor, network_type=network_type, is_training=is_training)


    def train_batch_generator(self, batch_size):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred1 = []
            y_pred2 = []
            y_pred3 = []


            for i in range(batch_size):
                if idx >= len(self.dataset_file_list):
                    idx = 0

                try:
                    tmp_x = cv2.imread(self.dataset_folder_path + self.dataset_file_list[idx])
                    tmp_x = cv2.resize(tmp_x, (self.input_width, self.input_height))
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y = self.read_target(self.label_folder_path + self.dataset_file_list[idx][:-3] + "txt")
                    x_batch.append(tmp_x)
                    y_pred1.append(tmp_y[0])
                    y_pred2.append(tmp_y[1])
                    y_pred3.append(tmp_y[2])

                except:
                    pass

                idx += 1

            yield (np.array(x_batch), [np.array(y_pred1), np.array(y_pred2), np.array(y_pred3)])
    
    



import os
import tensorflow as tf
from poda.layers.dense import *
from poda.layers.metrics import *
from poda.layers.optimizer import *
from poda.layers.activation import *
from poda.layers.convolutional import *
from poda.utils.visualize_training import *
from poda.transfer_learning.Vgg16_slim import *
from poda.transfer_learning.Vgg16 import VGG16
#from poda.transfer_learning.InceptionV4 import InceptionV4
from poda.preprocessing.GeneratorImage import GeneratorImage as generator

class ImageClassification(object):
    def __init__(self, classes, directory_dataset, batch_sizes=4, color_modes='rgb', image_sizes = (512,512,3), base_model='vgg_16', custom_architecture = False, dict_var_architecture={}):
        """[summary]
        
        Arguments:
            object {[type]} -- [description]
            classes {[type]} -- [description]
            directory_dataset {[type]} -- [description]
        
        Keyword Arguments:
            image_sizes {tuple} -- [description] (default: {(512,512,3)})
            base_model {str} -- [description] (default: {'vgg_16'})
            custom_architecture {bool} -- [description] (default: {False})
            dict_var_architecture {dict} -- [description] (default: {{}})
        """
        self.classes = classes
        self.dataset_folder_path = directory_dataset
        self.batch_size = batch_sizes
        self.color_mode = color_modes
        self.input_height = image_sizes[0]
        self.input_width = image_sizes[1]
        self.input_channel = image_sizes[2]
        self.image_size = image_sizes
        self.type_architecture = base_model
        self.custom_architecture = custom_architecture
        self.dict_var_architecture = dict_var_architecture

        self.input_tensor = tf.placeholder(tf.float32, shape=(self.batch_size, self.input_height, self.input_width, self.input_channel), name='input_tensor')
        self.output_tensor = tf.placeholder(tf.float32, (self.batch_size, self.classes),name='output_tensor')
    
    def create_model(self, dict_architecture):
        if self.type_architecture in "vgg_16_slim":
            if len(dict_architecture) > 0:
                num_block = dict_architecture['num_blocks']
                batch_normalization = dict_architecture['batch_normalizations']
                num_depthwise_layer = dict_architecture['num_depthwise_layers']
                num_dense_layer = dict_architecture['num_dense_layers']
                num_hidden_unit = dict_architecture['num_hidden_units']
                activation_dense = dict_architecture['activation_denses']
                dropout_rate = dict_architecture['dropout_rates']
                regularizer = dict_architecture['regularizers']
                scope = dict_architecture['scopes']
            else:
                num_block = 4
                batch_normalization = True
                num_depthwise_layer = 0
                num_dense_layer = 1
                num_hidden_unit = 512
                activation_dense = 'relu'
                dropout_rate = None
                regularizer = None
                scope = None

            if self.type_architecture == 'vgg_16':
                model = VGG16(input_tensor=self.input_tensor, num_blocks=num_block, classes=self.classes, batch_normalizations=batch_normalization, num_depthwise_layers=num_depthwise_layer,
                            num_dense_layers=num_dense_layer, num_hidden_units=num_hidden_unit, activation_denses=activation_dense, dropout_rates=dropout_rate, regularizers=regularizer, scopes=scope)
                non_logit, output, base_var_list, full_var_list = model.create_model()
            else:
                non_logit , output , base_var_list , full_var_list = vgg16(input_tensor=self.input_tensor,num_classes=self.classes,num_blocks=num_block,num_depthwise_layer=num_depthwise_layer,
                                                                       num_fully_connected_layer=num_dense_layer,num_hidden_unit=num_hidden_unit,activation_fully_connected=activation_dense,regularizers=regularizer)
        elif self.type_architecture in 'inception_v4_slim':
            if len(dict_architecture) > 0:
                inception_a = dict_architecture['n_inception_a']
                inception_b = dict_architecture['n_inception_b']
                inception_c = dict_architecture['n_inception_c']
                batch_normalization = dict_architecture['batch_normalizations']
                dropout_rate = dict_architecture['dropout_rates']
                regularizer = dict_architecture['regularizers']
                scope = dict_architecture['scopes']
            else:
                inception_a = 4
                inception_b = 7
                inception_c = 3
                batch_normalization = True
                dropout_rate = None
                regularizer = None
                scope = None

            if self.type_architecture == 'incetion_v4':
                model = InceptionV4(input_tensor=self.input_tensor, n_inception_a=inception_a, n_inception_b=inception_b, n_inception_c=inception_c, classes=self.classes, batch_normalizations = batch_normalization, 
                                    dropout_rates=dropout_rate, regularizers=regularizer, scopes=scope)
                non_logit, output, base_var_list, full_var_list = model.create_model()
                
            else:
                non_logit, output, base_var_list, full_var_list = inception_v4(inputs=self.input_tensor,  num_classes=self.classes,final_endpoint='Mixed_7d',is_training=batch_normalization,dropout_keep_prob=dropout_rate,
                                                                                num_depthwise_layer=None,regularizers=regularizer)

        return non_logit, output, base_var_list, full_var_list

    def train(self, epoch, output_model_path, dict_augmented_image={}, is_last_checkpoint=False, manual_split_dataset= False, use_pretrain = False, path_pretrained='', threshold_best_model=0.5, optimizers_name='adam', lr=0.0001):
        train_accuracy = []
        train_losess = []
        val_accuracy = []
        val_losess = []

        if self.custom_architecture:
            non_logit, output, base_var_list, full_var_list = self.create_model(dict_architecture=self.dict_var_architecture)
        else:
            non_logit, output, base_var_list, full_var_list = self.create_model(dict_architecture={})
            
        accuracy = calculate_accuracy_classification(output,self.output_tensor)

        cost = calculate_loss(input_tensor=non_logit, label=self.output_tensor, type_loss_function='softmax_crossentropy_mean')
        update_ops = tf.compat.v1.get_collection(tf.GraphKeys.UPDATE_OPS)
        with tf.control_dependencies(update_ops):
            optimizer = optimizers(optimizers_names=optimizers_name,learning_rates=lr).minimize(cost)

        init = tf.compat.v1.global_variables_initializer()

        if len(dict_augmented_image) > 0:
            rotation_degree = dict_augmented_image['rotation_degree']
            flip_horizontal = dict_augmented_image['flip_horizontal']
            flip_vertical = dict_augmented_image['flip_vertical']
            zoom_scale = dict_augmented_image['zoom_scale'] 
        else:
            rotation_degree = None
            flip_horizontal = False
            flip_vertical = False
            zoom_scale = None
            
        if manual_split_dataset:
            if len(dict_augmented_image) > 0:
                gen_train = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size, rotation_degrees=rotation_degree, flip_image_horizontal_status=flip_horizontal, 
                                                flip_image_vertical_status=flip_vertical, zoom_scales=zoom_scale)
                gen_val = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size, rotation_degrees=rotation_degree, flip_image_horizontal_status=flip_horizontal, 
                                                flip_image_vertical_status=flip_vertical, zoom_scales=zoom_scale)
            else:
                gen_train = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size)
                gen_val = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size)
                
            path_dataset_train = os.path.join(self.dataset_folder_path, 'train')
            path_dataset_val = os.path.join(self.dataset_folder_path, 'val')
            generator_train, num_train_data = gen_train.generate_from_directory_manual(directory=path_dataset_train)
            generator_val, num_val_data = gen_val.generate_from_directory_manual(directory=path_dataset_val)
        else:
            if len(dict_augmented_image) > 0:
                image_generator = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size, rotation_degrees=rotation_degree, flip_image_horizontal_status=flip_horizontal, 
                                                flip_image_vertical_status=flip_vertical, zoom_scales=zoom_scale)
            else:
                image_generator = generator(batch_sizes=self.batch_size, color_modes=self.color_mode, image_sizes=self.image_size)
                
            generator_train, generator_val, num_train_data, num_val_data = image_generator.generate_from_directory_auto(directory=self.dataset_folder_path, val_ratio=0.2)

        main_graph_saver = tf.compat.v1.train.Saver(base_var_list)
        output_saver = tf.compat.v1.train.Saver()

        output_saver_path = os.path.join(os.getcwd(),output_model_path,'full_model')
        if not os.path.exists(output_saver_path):
            os.makedirs(output_saver_path)

        main_graph_saver_path = os.path.join(os.getcwd(),output_model_path,'base_model')
        if not os.path.exists(main_graph_saver_path):
            os.makedirs(main_graph_saver_path)

        msg = "Epoch: {0:>6} loss: {1:>6.3} - acc: {2:>6.3} - val_loss: {3:>6.3} - val_acc: {4:>6.3} - {5}"
        
        train_iteration = int(num_train_data/self.batch_size)
        validation_iteration = int(num_val_data/self.batch_size)
        
        #with tf.Session() as sess:
        #    writer = tf.summary.FileWriter("output", sess.graph)
        #    sess.run(init)
        #    writer.close()

        counter_model = 1
        with tf.compat.v1.Session() as sess:
            sess.run(init)
            
            if is_last_checkpoint:
                output_saver.restore(sess,output_saver_path)

            if use_pretrain:
                main_graph_saver.restore(sess,main_graph_saver_path)            
            
            best_val_accuracy = threshold_best_model
            for i in range(epoch):
                print('Epoch '+str(i+1)+'/'+str(epoch))
                tmp_train_loss = []
                tmp_train_acc = []
                print("Step Train")
                for j in range(0,train_iteration):
                    sign = '--------------'
                    x_batch_train , y_batch_train = next(generator_train)
                    feed_dict_train = {self.input_tensor: x_batch_train , self.output_tensor: y_batch_train}
                    sess.run(optimizer,feed_dict=feed_dict_train)
                    y_prediction , train_loss = sess.run([output,cost], feed_dict=feed_dict_train)
                    train_acc = sess.run(accuracy, feed_dict=feed_dict_train)

                    print_progress_training(number_iteration=train_iteration, index_iteration=j, metrics_acc=train_acc, metrics_loss=train_loss, type_progress="train")
                    
                    tmp_train_loss.append(train_loss)
                    tmp_train_acc.append(train_acc)

                avg_train_loss = sum(tmp_train_loss)/(len(tmp_train_loss)+0.0001)
                avg_train_acc = sum(tmp_train_acc)/(len(tmp_train_acc)+0.0001)

                tmp_val_loss = []
                tmp_val_acc = []
                print("Step Validation")
                for k in range(0,validation_iteration):
                    x_batch_valid , y_batch_valid = next(generator_val)
                    feed_dict_valid = {self.input_tensor: x_batch_valid , self.output_tensor: y_batch_valid}
                    y_validation , val_loss = sess.run([output,cost], feed_dict=feed_dict_valid)
                    val_acc = sess.run(accuracy, feed_dict=feed_dict_valid)

                    print_progress_training(number_iteration=validation_iteration, index_iteration=k, metrics_acc=val_acc, metrics_loss=val_loss, type_progress="val")

                    tmp_val_loss.append(val_loss)
                    tmp_val_acc.append(val_acc)
                    
                avg_val_loss = sum(tmp_val_loss)/(len(tmp_val_loss)+0.0001)
                avg_val_acc = sum(tmp_val_acc)/(len(tmp_val_acc)+0.0001)
                    
                if avg_val_acc > best_val_accuracy:
                    base_model_path = os.path.join(main_graph_saver_path,"base_"+str(self.type_architecture))
                    full_model_path = os.path.join(output_saver_path,str(self.type_architecture))
                    main_graph_saver.save(sess=sess,save_path=base_model_path)
                    output_saver.save(sess=sess,save_path=full_model_path)
                    best_val_accuracy = avg_val_acc
                    sign = 'Found the Best '+str(counter_model)
                    counter_model+=1
                    
                train_accuracy.append(avg_train_acc)
                train_losess.append(avg_train_loss)
                val_accuracy.append(avg_val_acc)
                val_losess.append(avg_val_loss)

                print(msg.format(i, avg_train_loss, avg_train_acc, avg_val_loss, avg_val_acc , sign))
        return train_accuracy , train_losess, val_accuracy, val_losess





# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains model definitions for versions of the Oxford VGG network.
These model definitions were introduced in the following technical report:
  Very Deep Convolutional Networks For Large-Scale Image Recognition
  Karen Simonyan and Andrew Zisserman
  arXiv technical report, 2015
  PDF: http://arxiv.org/pdf/1409.1556.pdf
  ILSVRC 2014 Slides: http://www.robots.ox.ac.uk/~karen/pdf/ILSVRC_2014.pdf
  CC-BY-4.0
More information can be obtained from the VGG website:
www.robots.ox.ac.uk/~vgg/research/very_deep/
Usage:
  with slim.arg_scope(vgg.vgg_arg_scope()):
    outputs, end_points = vgg.vgg_a(inputs)
  with slim.arg_scope(vgg.vgg_arg_scope()):
    outputs, end_points = vgg.vgg_16(inputs)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from poda.layers.dense import *
from poda.layers.merge import *
from poda.layers.activation import *
from poda.layers.convolutional import *
from poda.utils import *
slim = tf.contrib.slim

def build_base_layer_model(input_tensor,num_blocks=5,scope='vgg_16'):
    with tf.variable_scope(scope, 'vgg_16', [input_tensor]) as sc:
        end_points_collection = sc.original_name_scope + '_end_points'
        # Collect outputs for conv2d, fully_connected and max_pool2d.
        with slim.arg_scope([slim.conv2d, slim.fully_connected, slim.max_pool2d],outputs_collections=end_points_collection):
            net = slim.repeat(input_tensor, 2, slim.conv2d, 64, [3, 3], scope='conv1')
            block_1 = slim.max_pool2d(net, [2, 2], scope='pool1')
            net = slim.repeat(block_1, 2, slim.conv2d, 128, [3, 3], scope='conv2')
            block_2 = slim.max_pool2d(net, [2, 2], scope='pool2')
            net = slim.repeat(block_2, 3, slim.conv2d, 256, [3, 3], scope='conv3')
            block_3 = slim.max_pool2d(net, [2, 2], scope='pool3')
            net = slim.repeat(block_3, 3, slim.conv2d, 512, [3, 3], scope='conv4')
            block_4 = slim.max_pool2d(net, [2, 2], scope='pool4')
            net = slim.repeat(block_4, 3, slim.conv2d, 512, [3, 3], scope='conv5')
            block_5 = slim.max_pool2d(net, [2, 2], scope='pool5')

            if num_blocks==1:
                net = block_1
            elif num_blocks==2:
                net = block_2
            elif num_blocks==3:
                net = block_3
            elif num_blocks==4:
                net = block_4
            elif num_blocks==5:
                net = block_5
            return net
            
def build_top_layer_model(base_layer,num_depthwise_layer=None,
                          num_fully_connected_layer=1,num_hidden_unit=512,
                          activation_fully_connected='relu',dropout_keep_prob=None,regularizers=None,num_classes=1000):
    previous_layer = base_layer
    if num_depthwise_layer!=None:
        num_depthwise_layer = num_depthwise_layer * 3
        for i in range(num_depthwise_layer):
            depth_wise_net = depthwise_convolution_2d(input_tensor=previous_layer,number_filters=1, 
                                                      kernel_sizes=(3,3), stride_sizes=(2,2), paddings='same',
                                                      activations='relu',names=str(i))
            previous_layer = depth_wise_net

    flatten_layer = flatten(input_tensor=previous_layer)

    if num_fully_connected_layer !=None:
        for j in range(num_fully_connected_layer):
            fully_connected_net = dense(input_tensor=flatten_layer,hidden_units=num_hidden_unit,
                                                  activations=activation_fully_connected,regularizers=regularizers,
                                                  scale=dropout_keep_prob,names=str(j))
            flatten_layer = fully_connected_net
    else:
        flatten_layer = flatten_layer

    non_logit = dense(input_tensor=flatten_layer,hidden_units=num_classes,activations=None)
    if num_classes > 2:
        output = softmax(input_tensor=non_logit, names='output')
    else:
        output = sigmoid(input_tensor=non_logit, names='output')
    return non_logit, output
    
def vgg16(input_tensor,num_classes=1000,num_blocks=5,num_depthwise_layer=None,num_fully_connected_layer=1,num_hidden_unit=512,activation_fully_connected=None,regularizers=None):
    net = build_base_layer_model(input_tensor=input_tensor,num_blocks=num_blocks)
    base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
    non_logit , output = build_top_layer_model(net,num_depthwise_layer=num_depthwise_layer,num_fully_connected_layer=num_fully_connected_layer,num_hidden_unit=num_hidden_unit,
                                               activation_fully_connected=activation_fully_connected,regularizers=regularizers,num_classes=num_classes)

    full_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)

    return non_logit , output , base_var_list , full_var_list

def vgg_arg_scope(self,weight_decay=0.0005):
    with slim.arg_scope([slim.conv2d, slim.fully_connected],
                            activation_fn=tf.nn.relu,
                            weights_regularizer=slim.l2_regularizer(weight_decay),
                            biases_initializer=tf.zeros_initializer()):
        with slim.arg_scope([slim.conv2d], padding='SAME') as arg_sc:
            return arg_sc


import tensorflow as tf
import os
import poda.tranfer_learning.Vgg16 as vgg

def convert_model_vggg16(model_path=''):
        input_string = tf.placeholder(tf.string, shape=[None], name='string_input')
        decode = lambda raw_byte_str: tf.image.resize_images(
                                            tf.cast(
                                                tf.image.decode_jpeg(raw_byte_str, channels=3, name='decoded_image'),
                                                    tf.uint8),
                                            [340, 340])

        input_images = tf.map_fn(decode, input_string, dtype=tf.float32)
        input_images = input_images / 255.0
        var_list , non_logit , top_layer_vgg16 = vgg.VGG16().create_pretrain_model_vgg(input_tensor=input_images,num_classes=1,num_blocks=4,is_last_checkpoint=False,path_last_checkpoint=None,is_training=False,num_depthwise_layer=1,num_fully_connected_layer=1,num_hidden_unit=512,activation_fully_connected='relu',regularizers=None)
        output_saver = tf.train.Saver()
        init = tf.initializers.global_variables()
        with tf.Session() as sess:
            writer = tf.summary.FileWriter("output", sess.graph)
            sess.run(init)
            writer.close()
            output_saver.restore(sess,model_path)
            builder = tf.saved_model.builder.SavedModelBuilder('model_extraction/')

            # Create aliase tensors
            # tensor_info_x: for input tensor
            # tensor_info_y: for output tensor
            tensor_info_x = tf.saved_model.utils.build_tensor_info(input_string )
            tensor_info_y = tf.saved_model.utils.build_tensor_info(top_layer_vgg16)

            # create prediction signature
            prediction_signature = tf.saved_model.signature_def_utils.build_signature_def(
                inputs={'input': tensor_info_x},
                outputs={'output': tensor_info_y},
                method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME)
            # So your input json is:
            # {"input": your features}
            # REMEMBER: the shape of your feature is without batch for single request

            # build frozen graph
            legacy_init_op = tf.group(tf.tables_initializer(), name='legacy_init_op')
            builder.add_meta_graph_and_variables(
            sess, [tf.saved_model.tag_constants.SERVING],
            signature_def_map={
                'predict_result':
                prediction_signature
            },
            legacy_init_op=legacy_init_op)
            builder.save()



import tensorflow as tf
from poda.layers.merge import *
from poda.layers.dense import *
from poda.layers.activation import *
from poda.layers.regularizer import *
from poda.layers.convolutional import *

class VGG16(object):
    def __init__(self, input_tensor, num_blocks=5, classes=1000, batch_normalizations = True, num_depthwise_layers=None, num_dense_layers=1, num_hidden_units=4096, activation_denses='relu', dropout_rates=None, regularizers=None, scopes=None):
        """[summary]
        
        Arguments:
            object {[type]} -- [description]
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            classes {int} -- [description] (default: {1000})
            batch_normalization {bool} -- [description] (default: {True})
        """
        self.input_tensor = input_tensor
        self.num_block = num_blocks
        self.classes = classes
        self.batch_normalization = batch_normalizations
        self.num_depthwise_layer = num_depthwise_layers
        self.num_dense_layer = num_dense_layers
        self.num_hidden_unit = num_hidden_units
        self.activation_dense = activation_denses
        self.dropout_rate = dropout_rates
        self.regularizer = regularizers
        self.scope = scopes

    def vgg_block(self, input_tensor, num_block=5, batch_normalization=True, scope=None):
        with tf.compat.v1.variable_scope(scope, 'vgg_16', [input_tensor]):
            with tf.compat.v1.variable_scope('Block_1'):
                conv_1 = convolution_2d(input_tensor=input_tensor, number_filters=64, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_2 = convolution_2d(input_tensor=conv_1, number_filters=64, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_3 = max_pool_2d(input_tensor=conv_2, pool_sizes=(2,2), stride_sizes=(2,2), paddings='valid', names=None)
            with tf.compat.v1.variable_scope('Block_2'):
                conv_4 = convolution_2d(input_tensor=conv_3, number_filters=128, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_5 = convolution_2d(input_tensor=conv_4, number_filters=128, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_6 = max_pool_2d(input_tensor=conv_5, pool_sizes=(2,2), stride_sizes=(2,2), paddings='valid', names=None)
            with tf.compat.v1.variable_scope('Block_3'):
                conv_7 = convolution_2d(input_tensor=conv_6, number_filters=256, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_8 = convolution_2d(input_tensor=conv_7, number_filters=256, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_9 = convolution_2d(input_tensor=conv_8, number_filters=256, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_10 = max_pool_2d(input_tensor=conv_9, pool_sizes=(2,2), stride_sizes=(2,2), paddings='valid', names=None)
            with tf.compat.v1.variable_scope('Block_4'):
                conv_11 = convolution_2d(input_tensor=conv_10, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_12 = convolution_2d(input_tensor=conv_11, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_13 = convolution_2d(input_tensor=conv_12, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_14 = max_pool_2d(input_tensor=conv_13, pool_sizes=(2,2), stride_sizes=(2,2), paddings='valid', names=None)
            with tf.compat.v1.variable_scope('Block_5'):
                conv_15 = convolution_2d(input_tensor=conv_14, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_16 = convolution_2d(input_tensor=conv_15, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_17 = convolution_2d(input_tensor=conv_16, number_filters=512, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
                conv_18 = max_pool_2d(input_tensor=conv_17, pool_sizes=(2,2), stride_sizes=(2,2), paddings='valid', names=None)

        if num_block==1:
            vgg_16 = conv_3
        elif num_block==2:
            vgg_16 = conv_6
        elif num_block==3:
            vgg_16 = conv_10
        elif num_block==4:
            vgg_16 = conv_14
        elif num_block==5:
            vgg_16 = conv_18
        else:
            vgg_16 = conv_18

        return vgg_16

    def create_model(self):
        number_filter = self.input_tensor.get_shape().as_list()[-1]

        vgg_base = self.vgg_block(input_tensor=self.input_tensor, num_block=self.num_block, batch_normalization=self.batch_normalization, scope=self.scope)

        base_var_list = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.GLOBAL_VARIABLES)
        
        with tf.variable_scope(self.scope, 'vgg_16', [vgg_base]):
            if self.num_depthwise_layer!=None or self.num_depthwise_layer>0:
                for j in range(0,self.num_depthwise_layer):
                    ##### FIX THIS TOMORROW
                    vgg_base = depthwise_convolution_2d(input_tensor=vgg_base, number_filters=number_filter, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None)
            else:
                flatten_layer = flatten(input_tensor=vgg_base, names='flatten')
                for i in range(0, self.num_dense_layer):
                    vgg_base = dense(input_tensor=flatten_layer, hidden_units=self.num_hidden_unit, activations=self.activation_dense, regularizers=self.regularizer, scale=self.dropout_rate)

            
            last_layer = flatten(input_tensor=vgg_base, names='flatten')

            full_var_list = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.GLOBAL_VARIABLES)

            non_logit = dense(input_tensor=last_layer, hidden_units=self.classes, names='output')

            if self.classes > 2:
                output = softmax(input_tensor=non_logit, names='output')
            else:
                output = sigmoid(input_tensor=non_logit, names='output')

        return non_logit, output, base_var_list, full_var_list


import tensorflow as tf

def add(input_tensor_1, input_tensor_2, names=None):
    """[summary]
    
    Arguments:
        input_tensor_1 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 1]
        input_tensor_2 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 2]
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Added tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'add'

    add_tensor = tf.math.add(x=input_tensor_1,y=input_tensor_2,name=names)
    return add_tensor

def concatenate(list_tensor, axis=-1, names=None):
    """[summary]
    
    Arguments:
        list_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A list tensor representing prelayer 1]
    
    Keyword Arguments:
        axis {int} -- [Dimension tensor] (default: {-1})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [Concatenate tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'concatenate'

    concat_tensor = tf.concat(values=list_tensor,axis=axis,name=names)
    return concat_tensor

def flatten(input_tensor, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
    
    Keyword Arguments:
        names {[type]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [Flatten tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'flatten'
        
    # Insert name future
    flatten_tensor = tf.contrib.layers.flatten(input_tensor)
    return flatten_tensor

def maximum(input_tensor_1, input_tensor_2, names=None):
    """[summary]
    
    Arguments:
        input_tensor_1 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 1]
        input_tensor_2 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 2]
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Maximum tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'maximum'

    maximum_tensor = tf.math.maximum(x=input_tensor_1,y=input_tensor_2,name=names)
    return maximum_tensor

def minimum(input_tensor_1, input_tensor_2, names=None):
    """[summary]
    
    Arguments:
        input_tensor_1 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 1]
        input_tensor_2 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 2]
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Minimum tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'minimum'

    minimum_tensor = tf.math.minimum(x=input_tensor_1,y=input_tensor_2,name=names)
    return minimum_tensor

def multiply(input_tensor_1, input_tensor_2, names=None):
    """[summary]
    
    Arguments:
        input_tensor_1 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 1]
        input_tensor_2 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 2]
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Multiplied tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'multipy'

    multiply_tensor = tf.math.multiply(x=input_tensor_1,y=input_tensor_2,name=names)
    return multiply_tensor

def substract(input_tensor_1, input_tensor_2, names=None):
    """[summary]
    
    Arguments:
        input_tensor_1 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 1]
        input_tensor_2 {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer 2]
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Substracted tensor]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'subtract'

    substract_tensor = tf.math.subtract(x=input_tensor_1,y=input_tensor_2,name=names)
    return substract_tensor




import tensorflow as tf
from poda.layers.activation import *
from poda.layers.regularizer import *

def dense(input_tensor, hidden_units, names=None, activations=None, regularizers=None, scale=0.2, bias_initializers='ZERO'):
    """[Create a new trainable Fully Connected layer]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
        hidden_units {[int]} -- [description]
        names {[str]} -- [Name of the layer] (default: {None})
    
    Keyword Arguments:
        activation {str} -- [description] (default: {'relu'})
        regularizers {[type]} -- [description] (default: {None})
        scale {float} -- [description] (default: {0.2})
    
    Returns:
        [Tensor] -- [A trainable tensor]
    """
    if names!=None:
        name = str(names)
    else:
        name = 'dense'

    weight = new_weights(shapes=[input_tensor.get_shape().as_list()[1], hidden_units], names=name)

    bias = new_biases(shapes=[input_tensor.get_shape().as_list()[0], hidden_units], names=name, bias_initializers=bias_initializers)

    layer = tf.matmul(input_tensor, weight)
    layer += bias

    if activations!=None:
        layer = define_activation_function(input_tensor=layer, activation_names=activations, names=name)
    else:
        layer = layer

    if regularizers=='both':
        layer = dropout(input_tensor=layer, dropout_rates=scale, names=name)
        layer = l2_regularization(input_tensor=layer, names=name)
    elif regularizers=='l1':
        layer = l1_regularization(input_tensor=layer)
    elif regularizers=='l2':
        layer = l2_regularization(input_tensor=layer)
    elif regularizers=='dropout':
        layer = dropout(input_tensor=layer, dropout_rates=scale, names=name)
    else:
        layer = layer

    return layer

def new_weights(shapes, names, random_types='truncated_normal', dtypes=tf.float32):
    """[Create a new trainable tensor as weight]
    
    Arguments:
        shapes {[Array of int]} -- [- example (convolution), [filter height, filter width, input channels, output channels]
                                    - example (fully connected), [num input, num output]]
        names {[str]} -- [A name for the operation (optional)]
    
    Keyword Arguments:
        random_types {str} -- [Type of random values] (default: {'truncated_normal'})
        dtypes {[float32]} -- [Data type of tensor] (default: {tf.float32})
    
    Returns:
        [float32] -- [A trainable tensor]
    """
    if random_types=='random_normal':
        initial_weight = tf.random.normal(shape=shapes,mean=0.0,stddev=1.0,dtype=dtypes,seed=None)
    elif random_types=='random_uniform':
        initial_weight = tf.random.uniform(shape=shapes,minval=0,maxval=None,dtype=dtypes,seed=None)
    else:
        initial_weight = tf.random.truncated_normal(shape=shapes,mean=0.0,stddev=1.0,dtype=dtypes,seed=None)
    return tf.Variable(initial_weight, dtype=dtypes, name='weight_'+names)

def new_biases(shapes, names, bias_initializers='NOTZERO', dtypes=tf.float32):
    """[Create a new trainable tensor as bias]
    
    Arguments:
        shapes {[int]} -- [Length of filter or node ]
        names {[str]} -- [A name for the operation (optional)]
    
    Keyword Arguments:
        dtypes {[float32]} -- [Data type of tensor] (default: {tf.float32})
    
    Returns:
        [float32] -- [A trainable tensor]
    """
    if bias_initializers=='ZERO':
        initial_bias = tf.zeros(shape=shapes, dtype=dtypes, name=names)
    else:
        initial_bias = tf.constant(0.05, shape=shapes, dtype=dtypes)

    return tf.Variable(initial_bias, dtype=dtypes, name='bias_'+names)



import numpy as np
import cv2
import imutils
import os
from sklearn.utils import shuffle

class GeneratorImage():
    def __init__(self, batch_sizes=4, color_modes='rgb', image_sizes=(512,512,3), rotation_degrees=None, flip_image_horizontal_status=False, flip_image_vertical_status=False, zoom_scales=None):
        """[summary]
        
        Arguments:
            batch_sizes {[type]} -- [description]
        
        Keyword Arguments:
            color_modes {str} -- [description] (default: {'rgb'})
            image_sizes {tuple} -- [description] (default: {(512,512,3)})
            rotation_degrees {[type]} -- [description] (default: {None})
            flip_image_horizontal_status {bool} -- [description] (default: {False})
            flip_image_vertical_status {bool} -- [description] (default: {False})
            zoom_scales {[type]} -- [description] (default: {None})
        """
        self.batch_size = batch_sizes
        self.color_mode = color_modes
        self.image_size = image_sizes
        self.rotation_degree = rotation_degrees
        self.flip_image_horizontal_status = flip_image_horizontal_status
        self.flip_image_vertical_status = flip_image_vertical_status
        self.zoom_scale = zoom_scales 

    def batch_generator(self, list_path_data, list_target_data):
        """[summary]
        
        Arguments:
            list_path_data {[type]} -- [description]
            list_target_data {[type]} -- [description]
        """
        counter_image = 0
        number_class = np.amax(list_target_data)
        while True:
            image_data_batch = []
            image_target_batch = []
            for i in range(self.batch_size):
                try:
                    number_augmentation = 1
                    image_file = list_path_data[counter_image]
                    if self.color_mode == 'grayscale':
                        image = cv2.imread(image_file,0)
                    else:
                        image = cv2.imread(image_file)
                    image = cv2.resize(image, (self.image_size[0], self.image_size[1]))
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                    image = image.astype(np.float32)
                    image = image/255.0

                    image_data_batch.append(image)
                    #image_target_batch.append(list_target_data[counter_image])
                    if self.rotation_degree!=None:
                        number_rotation = int(360/self.rotation_degree)
                        for j in range(number_rotation):
                            radian_rotation = self.rotation_degree
                            image_rotation = self.rotate_image(image, radian_rotation)
                            radian_rotation += self.rotation_degree
                            image_data_batch.append(image_rotation)
                            #image_target_batch.append(list_target_data[counter_image])
                            number_augmentation += 1
                    
                    if self.flip_image_horizontal_status==True:
                        image_flip = self.flip_image_horizontal(image)
                        image_data_batch.append(image_flip)
                        #image_target_batch.append(list_target_data[counter_image])
                        number_augmentation += 1

                    if self.flip_image_vertical_status==True:
                        image_flip = self.flip_image_vertical(image)
                        image_data_batch.append(image_flip)
                        #image_target_batch.append(list_target_data[counter_image])
                        number_augmentation += 1

                    if self.zoom_scale!=None:
                        image_zoom = self.zoom_image(image,self.zoom_scale)
                        image_data_batch.append(image_zoom)
                        #image_target_batch.append(list_target_data[counter_image])
                        number_augmentation += 1
                    
                    for i in range(0,number_augmentation):
                        tmp_target_batch = np.zeros(number_class+1).astype(np.float32)
                        tmp_target_batch[list_target_data[counter_image]-1] = 1
                        image_target_batch.append(tmp_target_batch)
                    
                    counter_image+=1
                    if counter_image >= len(list_path_data):
                        counter_image = 0
                except:
                    print("Failed to open "+str(list_path_data[counter_image]))

            image_data_batch = np.array(image_data_batch)            
            image_data_batch = image_data_batch.reshape((self.batch_size*number_augmentation,self.image_size[0],self.image_size[1],self.image_size[2])).astype(np.float32)
            image_target_batch = np.array(image_target_batch)            
            
            yield(image_data_batch,image_target_batch)

    def generate_from_directory_manual(self, directory):
        """[summary]
        
        Arguments:
            directory {[type]} -- [description]
        """
        list_path_image , list_target_image = self.get_list_image_and_label(path_dataset = directory)
        list_path_image , list_target_image = shuffle(list_path_image, list_target_image, random_state=0)
        return self.batch_generator(list_path_data=list_path_image, list_target_data=list_target_image) , self.get_num_data(list_path_image)

    def generate_from_directory_auto(self, directory, val_ratio=0.2):
        """[summary]
        
        Arguments:
            directory {[type]} -- [description]
        """
        list_path_image , list_target_image = self.get_list_image_and_label(path_dataset = directory)
        list_path_image , list_target_image = shuffle(list_path_image, list_target_image, random_state=0)
        split_size = 1 - val_ratio
        train_data, train_target, val_data, val_target = self.split_dataset(list_path_data=list_path_image, list_target_data=list_target_image, split_sizes=split_size)        
        train_generator = self.batch_generator(list_path_data=train_data, list_target_data=train_target)
        val_generator = self.batch_generator(list_path_data=val_data, list_target_data=val_target)
        num_train_data = self.get_num_data(train_data)
        num_val_data = self.get_num_data(val_data) 
        return train_generator , val_generator , num_train_data , num_val_data
    
    def get_num_data(self, list_path_data):
        """[summary]
        
        Arguments:
            list_path_data {[type]} -- [description]
        """
        return len(list_path_data)
        
    def flip_image_horizontal(self, image):
        """[summary]
        
        Arguments:
            image {[type]} -- [description]
        """
        return cv2.flip(image,0)

    def flip_image_vertical(self, image):
        """[summary]
        
        Arguments:
            image {[type]} -- [description]
        """
        return cv2.flip(image,1)
    
    def flip_image_both(self, image):
        """[summary]
        
        Arguments:
            image {[type]} -- [description]
        """
        return cv2.flip(image, -1)
    
    def get_list_image_and_label(self, path_dataset):
        """[summary]
        
        Arguments:
            path_dataset {[type]} -- [description]
        """
        dataset_path = os.path.join(os.getcwd(), path_dataset)
        list_dir_name = [d for d in os.listdir(dataset_path) if os.path.isdir(os.path.join(dataset_path,d))]
        list_dir_name.sort()
        list_path_file = []
        list_target_dataset = []
        for i in range(0,len(list_dir_name)):
            full_path_dir = os.path.join(dataset_path,list_dir_name[i])
            file_names = os.listdir(full_path_dir)
            for path_file in file_names:
                list_path_file.append(os.path.join(full_path_dir,path_file))
                list_target_dataset.append(i)
        list_path_dataset = np.array(list_path_file)
        list_target_dataset = np.array(list_target_dataset)
        list_target_dataset = np.reshape(list_target_dataset,len(list_path_dataset))
        return list_path_dataset, list_target_dataset

    def rotate_image(self, image, rotation_degrees=90):
        """[summary]
        
        Arguments:
            image {[type]} -- [description]
        
        Keyword Arguments:
            rotation_degrees {int} -- [description] (default: {90})
        """
        return imutils.rotate(image, rotation_degrees)

    def split_dataset(self, list_path_data, list_target_data, split_sizes=0.8):
        """[summary]
        
        Arguments:
            list_path_data {[type]} -- [description]
            list_target_data {[type]} -- [description]
        
        Keyword Arguments:
            split_sizes {float} -- [description] (default: {0.8})
        """
        limit = int(len(list_path_data) * split_sizes) + 1
        train_list_path_data = list_path_data[:limit]
        val_list_path_data = list_path_data[limit:]
        train_list_target_data = list_target_data[:limit]
        val_list_target_data = list_target_data[limit:]
        return train_list_path_data, train_list_target_data, val_list_path_data, val_list_target_data

    def zoom_image(self, image, zoom_scales=0.3):
        """[summary]
        
        Arguments:
            image {[type]} -- [description]
        
        Keyword Arguments:
            zoom_scales {float} -- [description] (default: {0.3})
        """
        height, width, channels = image.shape
        center_x,center_y=int(height/2),int(width/2)
        radius_x,radius_y= int(zoom_scales*height/100),int(zoom_scales*width/100)
        min_x,max_x=center_x-radius_x,center_x+radius_x
        min_y,max_y=center_y-radius_y,center_y+radius_y
        cropped = image[min_x:max_x, min_y:max_y]
        resized_cropped = cv2.resize(cropped, (width, height))
        return resized_cropped


import tensorflow as tf
from poda.layers.convolutional import *

class InceptionV4ResnetV2(object):
    def __init__(self, is_training = True):
        """[summary]
        
        Arguments:
            num_classes {[type]} -- [description]
        
        Keyword Arguments:
            input_tensor {[type]} -- [description] (default: {None})
            input_shape {tuple} -- [description] (default: {(None, 300, 300, 3)})
            learning_rate {float} -- [description] (default: {0.0001})
            is_training {bool} -- [description] (default: {True})
        """
        self.is_training = is_training


    def conv_block(self, 
                   inputs, 
                   filters, 
                   kernel_size, 
                   strides=(2,2), 
                   padding='VALID',  
                   dropout_rate=0.2, 
                   activation='relu',
                   batch_normalization=True):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
            filters {[type]} -- [description]
            kernel_size {[type]} -- [description]
        
        Keyword Arguments:
            strides {tuple} -- [description] (default: {(2,2)})
            padding {str} -- [description] (default: {'valid'})
            batch_normalization {bool} -- [description] (default: {True})
            dropout_rate {float} -- [description] (default: {0.15})
            activation {str} -- [description] (default: {'relu'})
            is_training {bool} -- [description] (default: {True})
        
        Returns:
            [type] -- [description]
        """
        conv = convolution_2d(input_tensor=inputs, number_filters=filters, kernel_sizes=kernel_size, 
                              stride_sizes=strides, paddings=padding, activations=activation)
        
        conv = batch_normalization(input_tensor=conv, is_trainable=batch_normalization)
        conv = dropout(input_tensor=conv, dropout_rates=dropout_rate)
        return conv


    def stem_block(self, input_tensor, batch_normalization=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        conv_1 = self.conv_block(inputs=input_tensor, filters=32, kernel_size=(3,3), strides=(2, 2), batch_normalization=batch_normalization)        
        conv_2 = self.conv_block(inputs=conv_1, filters=32, kernel_size=(3,3), strides=(1,1), batch_normalization=batch_normalization)
        
        conv_3 = self.conv_block(inputs=conv_2, filters=64, kernel_size=(3,3), padding='same', strides=(1, 1), batch_normalization=batch_normalization)
        conv_4 = self.conv_block(inputs=conv_3, filters=96, kernel_size=(3,3), strides=(2,2), batch_normalization=batch_normalization)
        
        max_pool_1 = max_pool_2d(input_tensor=conv_3, pool_sizes=(3,3), stride_sizes=(2,2))
        concat_1 = tf.concat([conv_4, max_pool_1], -1)

        conv_5 = self.conv_block(inputs=concat_1, filters=64, kernel_size=(3,3), strides=(1,1), batch_normalization=batch_normalization)
        conv_6 = self.conv_block(inputs=conv_5, filters=64, kernel_size=(7,1), padding='same', strides=(1, 1), batch_normalization=batch_normalization)
        
        conv_7 = self.conv_block(inputs=conv_6, filters=64, kernel_size=(1,7), padding='same', strides=(1, 1), batch_normalization=batch_normalization)
        conv_8 = self.conv_block(inputs=conv_7, filters=96, kernel_size=(3,3), padding='same', strides=(1, 1), batch_normalization=batch_normalization)
        
        conv_9 = self.conv_block(inputs=concat_1, filters=64, kernel_size=(1,1), strides=(1,1), batch_normalization=batch_normalization)
        conv_10 = self.conv_block(inputs=conv_9, filters=96, kernel_size=(3,3), strides=(1,1), batch_normalization=batch_normalization)
        
        concat_2 = tf.concat([conv_8, conv_10], -1)
        max_pool_2 = max_pool_2d(inputs=concat_2, pool_sizes=(3,3), stride_sizes=(2,2))
        
        conv_11 = self.conv_block(inputs=concat_2, filters=192, kernel_size=(3,3), strides=(2,2), batch_normalization=batch_normalization)
        concat_3 = tf.concat([max_pool_2, conv_11], -1)
        return concat_3
    

    def inception_resnet_a(self, input_tensor, drop_out=0.80, batch_normalization=True):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
        
        Keyword Arguments:
            drop_out {float} -- [description] (default: {0.85})
            activation {str} -- [description] (default: {'NONE'})
            is_training {bool} -- [description] (default: {True})
            use_bias {bool} -- [description] (default: {True})
            use_batchnorm {bool} -- [description] (default: {True})
        """
        
        conv_1 = self.conv_block(inputs=input_tensor, filters=64, kernel_size=(1,1) , strides=(1,1), batch_normalization=batch_normalization)
        conv_2 = self.conv_block(inputs=conv_1, filters=96, kernel_size=(3,3) , strides=(1,1), batch_normalization=batch_normalization)
        conv_3 = self.conv_block(inputs=conv_2, filters=96, kernel_size=(3,3) , strides=(1,1), batch_normalization=batch_normalization)
    
        conv_4 = self.conv_block(inputs=input_tensor, filters=64, kernel_size=(1,1) , strides=(1,1), batch_normalization=batch_normalization)
        conv_5 = self.conv_block(inputs=conv_4, filters=96, kernel_size=(3,3) , strides=(1,1), batch_normalization=batch_normalization)

        conv_6 = self.conv_block(inputs=input_tensor, filters=96, kernel_size=(1,1) , strides=(1,1), batch_normalization=batch_normalization)

        
        conv_7 = self.conv_block(inputs=conv_2, filters=96, kernel_size=(3,3) , strides=(1,1), batch_normalization=batch_normalization)
        input_depth = conv_right1.get_shape().as_list()[-1]
        conv_right2, _ = new_conv2d_layer(input=conv_right1, 
                                    filter_shape=[3, 3, input_depth, 48], 
                                    name='inres_a_conv_r2', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = conv_right2.get_shape().as_list()[-1]
        conv_right3, _ = new_conv2d_layer(input=conv_right2, 
                                    filter_shape=[3, 3, input_depth, 32], 
                                    name='inres_a_conv_r3', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_mid1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[1, 1, input_depth, 32], 
                                    name='inres_a_conv_m1', 
                                    dropout_val=drop_out, 
                                    activation=activation,
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = conv_mid1.get_shape().as_list()[-1]
        conv_mid2, _ = new_conv2d_layer(input=conv_mid1, 
                                    filter_shape=[3, 3, input_depth, 32], 
                                    name='inres_a_conv_m2', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_left1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[1, 1, input_depth, 32], 
                                    name='inres_a_conv_l1', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        concat_conv = tf.concat([conv_right3, conv_mid2, conv_left1], -1)
    
        input_depth = concat_conv.get_shape().as_list()[-1]
        output_depth = inputs.get_shape().as_list()[-1]
        conv_mixed, _ = new_conv2d_layer(input=concat_conv, 
                                    filter_shape=[1, 1, input_depth, output_depth], 
                                    name='inres_a_conv_concat', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        final_conv = inputs + conv_mixed
        return final_conv
        #return tf.nn.leaky_relu(final_conv)
    
    
    # Reduction A
    def reduction_a(self, inputs, 
                    drop_out=0.80, 
                    activation='NONE', 
                    use_bias=True, 
                    use_batchnorm=True):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
        
        Keyword Arguments:
            drop_out {float} -- [description] (default: {0.85})
            activation {str} -- [description] (default: {'NONE'})
            is_training {bool} -- [description] (default: {True})
            use_bias {bool} -- [description] (default: {True})
            use_batchnorm {bool} -- [description] (default: {True})
        """
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_right1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[1, 1, input_depth, 256], 
                                    name='reduc_a_conv_r1', 
                                    dropout_val=drop_out, 
                                    activation=activation,
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = conv_right1.get_shape().as_list()[-1]
        conv_right2, _ = new_conv2d_layer(input=conv_right1, 
                                    filter_shape=[3, 3, input_depth, 256], 
                                    name='reduc_a_conv_r2', 
                                    dropout_val=drop_out, 
                                    activation=activation,
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
        
        input_depth = conv_right2.get_shape().as_list()[-1]
        conv_right3, _ = new_conv2d_layer(input=conv_right2, 
                                    filter_shape=[3, 3, input_depth, 384], 
                                    name='reduc_a_conv_r3', 
                                    dropout_val=drop_out, 
                                    activation=activation,
                                    strides=[1, 2, 2, 1],
                                    padding='VALID',
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_mid1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[3, 3, input_depth, 384], 
                                    name='reduc_a_conv_m1', 
                                    dropout_val=drop_out, 
                                    activation=activation,
                                    strides=[1, 2, 2, 1],
                                    padding='VALID',
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
        
        """
        max_pool = tf.nn.max_pool(value=inputs,
                                ksize=[1, 3, 3, 1],
                                strides=[1, 1, 1, 15],
                                padding='VALID',
                                name='reduc_a_conv_mp')
        """
        return tf.concat([conv_right3, conv_mid1], -1, name='hellloooooo')
    
    
    # Inception ResNet B
    def inception_resnet_b(self, inputs, 
                            drop_out=0.80, 
                            activation='NONE', 
                            use_bias=True, 
                            use_batchnorm=True):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
        
        Keyword Arguments:
            drop_out {float} -- [description] (default: {0.85})
            activation {str} -- [description] (default: {'NONE'})
            is_training {bool} -- [description] (default: {True})
            use_bias {bool} -- [description] (default: {True})
            use_batchnorm {bool} -- [description] (default: {True})
        """
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_right1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[1, 1, input_depth, 128], 
                                    name='inres_a_conv_r1', 
                                    dropout_val=drop_out, 
                                    activation=activation,  
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = conv_right1.get_shape().as_list()[-1]
        conv_right2, _ = new_conv2d_layer(input=conv_right1, 
                                    filter_shape=[1, 7, input_depth, 160], 
                                    name='inres_a_conv_r2', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = conv_right2.get_shape().as_list()[-1]
        conv_right3, _ = new_conv2d_layer(input=conv_right2, 
                                    filter_shape=[7, 1, input_depth, 192], 
                                    name='inres_a_conv_r3', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        input_depth = inputs.get_shape().as_list()[-1]
        conv_mid1, _ = new_conv2d_layer(input=inputs, 
                                    filter_shape=[1, 1, input_depth, 192], 
                                    name='inres_a_conv_m1', 
                                    dropout_val=drop_out, 
                                    activation=activation, 
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        concat_conv = tf.concat([conv_right3, conv_mid1], -1)
        
        input_depth = concat_conv.get_shape().as_list()[-1]
        output_depth = inputs.get_shape().as_list()[-1]
        conv_mixed, _ = new_conv2d_layer(input=concat_conv, 
                                    filter_shape=[1, 1, input_depth, output_depth], 
                                    name='inres_a_conv_mx', 
                                    dropout_val=drop_out, 
                                    activation=activation,   
                                    is_training=self.is_training,
                                    use_bias=use_bias,
                                    use_batchnorm=use_batchnorm)
    
        final_conv = inputs + conv_mixed
        return final_conv
        #return tf.nn.leaky_relu(final_conv)


    def create_base_model(self, input=None):
        """[summary]
        
        Arguments:
            classes {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        with tf.variable_scope("stem"):
            net = self.stem_block(input)
        print (net, "===================>>>")
        
        with tf.variable_scope("inception_resnet_a"):
            for i in range(5):
                net = self.inception_resnet_a(inputs=net)
        print (net, "===================>>>")
        with tf.variable_scope("reduction_a"):
            net = self.reduction_a(inputs=net)
        print(net, "===================>>>")
        with tf.variable_scope("inception_resnet_b"):
            for i in range(10):
                 net = self.inception_resnet_b(inputs=net)
        print (net, "===================>>>")
        return net



import tensorflow as tf

def crelu(input_tensor, names=None):
    """[Concatenates a ReLU which selects only the positive part of the activation
        with a ReLU which selects only the negative part of the activation 
        Source: https://arxiv.org/abs/1603.05201]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing preactivation values]
        name {[str]} -- [A name for the operation (optional)]  (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation crelu]
    """
    return tf.nn.crelu(features=input_tensor,name='activation_crelu_'+names,axis=-1)

def define_activation_function(input_tensor, activation_names, names=None):
    """[Determine the type of activation function]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing preactivation values]
        activation_names {[str]} -- [Type of activation function]
    
    Keyword Arguments:
        names {[str]} -- [A name for the operation (optional)] (default: {None})
        
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with spesific activation function]
    """
    if activation_names=="crelu":
        output_tensor = tf.nn.crelu(features=input_tensor, name='activation_'+names, axis=-1)
    elif activation_names=="elu":
        output_tensor = tf.nn.elu(features=input_tensor, name='activation_'+names)
    elif activation_names=="leakyRelu":
        alpha=tf.constant(0.2, dtype=tf.float32)
        output_tensor = tf.nn.leaky_relu(features=input_tensor, alpha=alpha, name='activation_'+names)
    elif activation_names=="relu":
        output_tensor = tf.nn.relu(features=input_tensor, name='activation_'+names)
    elif activation_names=="relu6":
        output_tensor = tf.nn.relu6(features=input_tensor, name='activation_'+names)
    elif activation_names=="selu":
        output_tensor = tf.nn.selu(features=input_tensor, name='activation_'+names)
    elif activation_names=="sigmoid":
        output_tensor = tf.nn.sigmoid(x=input_tensor, name='activation_'+names)
    elif activation_names=="softmax":
        output_tensor = tf.nn.softmax(logits=input_tensor, axis=None, name='activation_'+names, dim=None)
    elif activation_names=="softsign":
        output_tensor = tf.nn.softsign(features=input_tensor, name='activation_'+names)
    elif activation_names=="tanh":
        output_tensor = tf.nn.tanh(x=input_tensor, name='activation_'+names)
    elif activation_names==None:
        output_tensor = input_tensor
    return output_tensor

def elu(input_tensor, names=None):
    """[Computes exponential linear: exp(features) - 1 if < 0, features otherwise
        Source: https://arxiv.org/abs/1511.07289]
    
    Arguments:
        input_tensor {[half, bfloat16, float32, float64]} -- [A Tensor representing preactivation values]
        name {[str]} -- [A name for the operation (optional)]  (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation elu]
    """
    return tf.nn.elu(features=input_tensor,name='activation_elu_'+names)

def leaky_relu(input_tensor, names=None, alpha=0.2):
    """[Compute the Leaky ReLU activation function
        Source: https://ai.stanford.edu/~amaas/papers/relu_hybrid_icml2013_final.pdf]
    
    Arguments:
        input_tensor {[float16, float32, float64, int32, int64]} -- [A Tensor representing preactivation values]
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Keyword Arguments:
        alpha {float} -- [Slope of the activation function at x < 0] (default: {0.2})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation leaky relu]
    """
    return tf.nn.leaky_relu(features=input_tensor,alpha=alpha,name='activation_lrelu_'+names)

def relu(input_tensor, names=None):
    """[Computes rectified linear: max(features, 0)]
    
    Arguments:
        input_tensor {[float32, float64, int32, uint8, int16, int8, int64, bfloat16, uint16, half, uint32, uint64, qint8]} 
                        -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation relu]
    """
    return tf.nn.relu(features=input_tensor,name='activation_relu_'+names)

def relu6(input_tensor, names=None):
    """[Computes Rectified Linear 6: min(max(features, 0), 6)
        Source: http://www.cs.utoronto.ca/~kriz/conv-cifar10-aug2010.pdf]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation relu6]
    """
    return tf.nn.relu6(features=input_tensor,name='activation_relu6_'+names)

def selu(input_tensor, names=None):
    """[Computes scaled exponential linear: scale * alpha * (exp(features) - 1)
        if < 0, scale * features otherwise]
    
    Arguments:
        input_tensor {[half, bfloat16, float32, float64]} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation selu]
    """
    return tf.nn.selu(features=input_tensor,name='activation_selu_'+names)

def sigmoid(input_tensor , names=None):
    """[Computes sigmoid of x element-wise.
        Specifically, y = 1 / (1 + exp(-x))]
    
    Arguments:
        input_tensor {[float16, float32, float64, complex64, or complex128]} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation sigmoid]
    """
    return tf.nn.sigmoid(x=input_tensor,name='activation_sigmoid_'+names)

def softmax(input_tensor, names=None, axis=None):
    """[Computes softmax activations]
    
    Arguments:
        input_tensor {[half, float32, float64]} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        axis {[int]} -- [The dimension softmax would be performed on. The default is -1 which indicates the last dimension.] (default: {None})
        name {[str]} -- [A name for the operation (optional)] (default: {None})
        dim {[int]} -- [Deprecated alias for axis] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation softmax]
    """
    return tf.nn.softmax(logits=input_tensor,axis=axis,name='activation_softmax_'+names)

def softsign(input_tensor, names=None):
    """[Computes softsign: features / (abs(features) + 1).]
    
    Arguments:
        input_tensor {[half, bfloat16, float32, float64]} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation softsign]
    """
    return tf.nn.softsign(features=input_tensor,name='activation_softsign_'+names)

def tanh(input_tensor, names=None):
    """[Computes hyperbolic tangent of x element-wise]
    
    Arguments:
        input_tensor {[bfloat16, half, float32, float64, complex64]} -- [A Tensor representing preactivation values]
    
    Keyword Arguments:
        name {[str]} -- [A name for the operation (optional)] (default: {None})
    
    Returns:
        [Tensor] -- [A tensor dtype tf.float32 with activation tanh]
    """
    return tf.nn.tanh(x=input_tensor,name='activation_tanh_'+names)


def print_progress_training(number_iteration=0, index_iteration=0, metrics_acc=0, metrics_loss=0, type_progress=""):
    split_iteration = int((number_iteration*5)/100)
    if split_iteration == 0:
        split_iteration = 1
    progress_bar = "--------------------"
    for i in range(0,index_iteration):
        if i % split_iteration == 0:
            progress_bar = ">" + progress_bar[:-1]
        else:
            progress_bar = progress_bar

    print(str(index_iteration+1)+"/"+str(number_iteration)+"    ["+str(progress_bar)+"]  "+str(type_progress)+"_acc: "+str(metrics_acc)+"       "+str(type_progress)+"_loss: "+str(metrics_loss))



import tensorflow as tf

def adadelta(learning_rates=0.0001, rho=0.95, epsilon=1e-08, use_lockings=False, names=None):
    """[summary]
    
    Keyword Arguments:
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
        rho {float} -- [Value of learning rho] (default: {0.999})
        epsilons {[type]} -- [Value of learning epsilons] (default: {1e-08})
        use_lockings {bool} -- [State of locking] (default: {False})
        names {[type]} -- [Name of the optimizer] (default: {None})

    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    if names!=None:
        names = str(names)
    else:
        names= ''
    
    optimizer = tf.compat.v1.train.AdadeltaOptimizer(learning_rate=learning_rates,rho=rho,epsilon=epsilon,use_locking=use_lockings,name=names)
    return optimizer

def adagrad(learning_rates=0.0001, initial_accumulator_value=0.1, use_lockings=False, names=None):
    """[summary]
    
    Keyword Arguments:
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
        initial_accumulator_value {float} -- [Value of learning rho] (default: {0.999})
        use_lockings {bool} -- [State of locking] (default: {False})
        names {[type]} -- [Name of the optimizer] (default: {None})

    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    if names!=None:
        names = str(names)
    else:
        names= ''
    
    optimizer = tf.compat.v1.train.AdagradOptimizer(learning_rate=learning_rates,initial_accumulator_value=initial_accumulator_value,use_locking=use_lockings,name=names)
    return optimizer

def adam(learning_rates=0.0001, beta1=0.9, beta2=0.999, epsilon=1e-08, use_lockings=False, names=None):
    """[summary]
    
    Keyword Arguments:
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
        beta1 {float} -- [Value of learning beta1] (default: {0.9})
        beta2 {float} -- [Value of learning beta2] (default: {0.999})
        epsilons {[type]} -- [Value of learning epsilons] (default: {1e-08})
        use_lockings {bool} -- [State of locking] (default: {False})
        names {[type]} -- [Name of the optimizer] (default: {None})

    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    if names!=None:
        names = str(names)
    else:
        names= ''

    optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=learning_rates,beta1=beta1,beta2=beta2,epsilon=epsilon,use_locking=use_lockings,name=names)
    return optimizer

def optimizers(optimizers_names='adam', learning_rates=0.0001):
    """[summary]
    
    Keyword Arguments:
        optimizers_names {str} -- [Type of optimizer] (default: {'adam'})
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
    
    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    optimizers_function = None
    if optimizers_names=='adam':
        optimizers_function = tf.compat.v1.train.AdamOptimizer(learning_rate=learning_rates,beta1=0.9,beta2=0.999,epsilon=1e-08,use_locking=False,name='Adam')
    elif optimizers_names=='adadelta':
        optimizers_function = tf.compat.v1.train.AdadeltaOptimizer(learning_rate=learning_rates,rho=0.95,epsilon=1e-08,use_locking=False,name='Adadelta')
    elif optimizers_names=='adagrad':
        optimizers_function = tf.compat.v1.train.AdagradOptimizer(learning_rate=learning_rates,initial_accumulator_value=0.1,use_locking=False,name='Adagrad')
    elif optimizers_names=='sgd':
        optimizers_function = tf.compat.v1.train.GradientDescentOptimizer(learning_rate=learning_rates,use_locking=False,name='GradientDescent')
    elif optimizers_names=='rmsprop':
        optimizers_function = tf.compat.v1.train.RMSPropOptimizer(learning_rate=learning_rates,decay=0.9,momentum=0.0,epsilon=1e-10,use_locking=False,centered=False,name='RMSProp')
    return optimizers_function

def rmsprop(learning_rates=0.0001, decay=0.9, momentum=0.0, epsilon=1e-10, use_lockings=False, centered=False, names=None):
    """[summary]
    
    Keyword Arguments:
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
        rho {float} -- [Value of learning rho] (default: {0.999})
        epsilons {[type]} -- [Value of learning epsilons] (default: {1e-08})
        use_lockings {bool} -- [State of locking] (default: {False})
        names {[type]} -- [Name of the optimizer] (default: {None})

    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    if names!=None:
        names = str(names)
    else:
        names= ''
    
    optimizer = tf.compat.v1.train.RMSPropOptimizer(learning_rate=learning_rates,decay=decay,momentum=momentum,epsilon=epsilon,use_locking=use_lockings,centered=centered,name=names)
    return optimizer

def sgd(learning_rates=0.0001, use_lockings=False, names=None):
    """[summary]
    
    Keyword Arguments:
        learning_rates {float} -- [Value of learning rate] (default: {0.0001})
        use_lockings {bool} -- [State of locking] (default: {False})
        names {[type]} -- [Name of the optimizer] (default: {None})

    Returns:
        [Optimizer] -- [Optimizer for training graph]
    """
    if names!=None:
        names = str(names)
    else:
        names= ''

    optimizer = tf.compat.v1.train.GradientDescentOptimizer(learning_rate=learning_rates,use_locking=use_lockings,name=names)
    return optimizer


import tensorflow as tf
from poda.layers.merge import *
from poda.layers.dense import *
from poda.layers.merge import *
from poda.layers.activation import *
from poda.layers.regularizer import *
from poda.layers.convolutional import *


class InceptionResnetV2(object):
    def __init__(self, input_tensor, n_inception_a=5, n_inception_b=10, n_inception_c=5, classes=1000, batch_normalizations = True, activation_denses='relu', dropout_rates=None, regularizers=None, scopes=None):
        """[summary]
        
        Arguments:
            num_classes {[type]} -- [description]
        
        Keyword Arguments:
            input_tensor {[type]} -- [description] (default: {None})
            input_shape {tuple} -- [description] (default: {(None, 300, 300, 3)})
            learning_rate {float} -- [description] (default: {0.0001})
            batch_normalizations {bool} -- [description] (default: {True})
        """
        self.input_tensor = input_tensor
        self.classes = classes
        self.n_inception_a = n_inception_a
        self.n_inception_b = n_inception_b
        self.n_inception_c = n_inception_c
        self.batch_normalizations = batch_normalizations
        self.activation_dense = activation_denses
        self.dropout_rate = dropout_rates
        self.regularizer = regularizers
        self.scope = scopes        


    def conv_block(self, inputs, filters, kernel_size, strides=(2,2), padding='VALID',  dropout_rate=0.2, activation=None, name=None, use_batch_norm=True):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
            filters {[type]} -- [description]
            kernel_size {[type]} -- [description]
        
        Keyword Arguments:
            strides {tuple} -- [description] (default: {(2,2)})
            padding {str} -- [description] (default: {'VALID'})
            dropout_rate {float} -- [description] (default: {0.2})
            activation {[type]} -- [description] (default: {None})
            name {[type]} -- [description] (default: {None})
            use_batch_norm {bool} -- [description] (default: {True})
        """
        conv = convolution_2d(input_tensor=inputs, number_filters=filters, kernel_sizes=kernel_size, stride_sizes=strides, paddings=padding, activations=activation, names=name)
        
        conv = batch_normalization(input_tensor=conv, is_trainable=use_batch_norm)
        conv = dropout(input_tensor=conv, dropout_rates=dropout_rate)
        return conv

    # Stem Block
    def stem_block(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {[type]} -- [description] (default: {None})
        """
        with tf.compat.v1.variable_scope(scope, 'StemBlock', [input_tensor], reuse=reuse):
            conv_1 = self.conv_block(inputs=input_tensor, filters=32, kernel_size=(3,3), padding='VALID', strides=(2, 2), use_batch_norm=batch_normalization, name='0_3x3')
            conv_2 = self.conv_block(inputs=conv_1, filters=32, kernel_size=(3,3), padding='VALID', strides=(1,1), use_batch_norm=batch_normalization, name='0_3x3')
            conv_3 = self.conv_block(inputs=conv_2, filters=64, kernel_size=(3,3), padding='SAME', strides=(1, 1), use_batch_norm=batch_normalization, name='0_3x3')
            with tf.compat.v1.variable_scope('Branch_0'):
                max_pool_1 = max_pool_2d(input_tensor=conv_3, pool_sizes=(3,3), stride_sizes=(2,2), paddings='VALID', names='0a_3x3')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_4 = self.conv_block(inputs=conv_3, filters=96, kernel_size=(3,3), padding='VALID', strides=(2,2), use_batch_norm=batch_normalization, name='0b_3x3')
            
            with tf.compat.v1.variable_scope('Concatenate_1'):
                concat_1 = concatenate(list_tensor=[conv_4,max_pool_1], axis=-1, names='conv_3x3_maxpool_3x3')

            with tf.compat.v1.variable_scope('Branch_3'):
                conv_5 = self.conv_block(inputs=concat_1, filters=64, kernel_size=(1,1), padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0a_1x1')
                conv_6 = self.conv_block(inputs=conv_5, filters=96, kernel_size=(3,3), padding='VALID', strides=(1,1), use_batch_norm=batch_normalization, name='1a_3x3')
            with tf.compat.v1.variable_scope('Branch_4'):
                conv_7 = self.conv_block(inputs=concat_1, filters=64, kernel_size=(1,1), padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0b_1x1')
                conv_8 = self.conv_block(inputs=conv_7, filters=64, kernel_size=(7,1), padding='SAME', strides=(1, 1), use_batch_norm=batch_normalization, name='1b_7x1')
                conv_9 = self.conv_block(inputs=conv_8, filters=64, kernel_size=(1,7), padding='SAME', strides=(1, 1), use_batch_norm=batch_normalization, name='2b_1x7')
                conv_10 = self.conv_block(inputs=conv_9, filters=96, kernel_size=(3,3), padding='VALID', strides=(1, 1), use_batch_norm=batch_normalization, name='3b_3x3')
            
            with tf.compat.v1.variable_scope('Concatenate_2'):
                concat_2 = concatenate(list_tensor=[conv_6,conv_10], axis=-1, names='conv_3x3_conv_3x3')

            with tf.compat.v1.variable_scope('Branch_5'):
                conv_11 = self.conv_block(inputs=concat_2, filters=192, kernel_size=(3,3), padding='VALID', strides=(2,2), use_batch_norm=batch_normalization, name='0a_3x3')
            with tf.compat.v1.variable_scope('Branch_6'):
                max_pool_2 = max_pool_2d(inputs=concat_2, pool_sizes=(3,3), stride_sizes=(2,2), paddings='VALID', names='0b_3x3')
            
            with tf.compat.v1.variable_scope('Concatenate_3'):
            concat_3 = concatenate(list_tensor=[max_pool_2,conv_11], axis=-1, names='maxpool_3x3_conv_3x3')
        return concat_3
    
    # Inception A
    def inception_resnet_a(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {bool} -- [description] (default: {True})
        """
        with tf.compat.v1.variable_scope(scope, 'BlockInceptionResnetA', [input_tensor], reuse=reuse):
            with tf.compat.v1.variable_scope('Relu_1'):
                relu_1 = relu(input_tensor=input_tensor, names='input_inception_resnet_a')
            with tf.compat.v1.variable_scope('Branch_0'):
                conv_1 = self.conv_block(inputs=relu_1, filters=32, kernel_size=(1,1), padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0a_1x1')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_2 = self.conv_block(inputs=relu_1, filters=32, kernel_size=(1,1), padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0b_1x1')
                conv_3 = self.conv_block(inputs=conv_2, filters=32, kernel_size=(3,3), padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1b_3x3')
            with tf.compat.v1.variable_scope('Branch_2'):
                conv_4 = self.conv_block(inputs=relu_1, filters=32, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0c_1x1')
                conv_5 = self.conv_block(inputs=conv_4, filters=48, kernel_size=(3,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1c_3x3')
                conv_6 = self.conv_block(inputs=conv_5, filters=64, kernel_size=(3,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='2c_3x3')
            
            with tf.compat.v1.variable_scope('Concatenate_1'):
                concat_1 = concatenate(list_tensor=[conv_1,conv_3,conv_6], axis=-1, names='conv_1x1_conv_3x3_conv_3x3')
            
            with tf.compat.v1.variable_scope('Branch_1c'):
                conv_7 = self.conv_block(inputs=concat_1, filters=384, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0aa_1x1')
            with tf.compat.v1.variable_scope('Add_1'):
                add_1 = add(input_tensor_1=relu_1, input_tensor_2=conv_7, names='input_conv_1x1')

            with tf.compat.v1.variable_scope('Relu_2'):
                relu_2 = relu(input_tensor=add_1, names='output_inception_resnet_a')
        return relu_2    
    
    # Reduction A
    def reduction_a(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {[type]} -- [description] (default: {None})
        """
        with tf.compat.v1.variable_scope(scope, 'BlockReductionA', [input_tensor], reuse=reuse):
            with tf.compat.v1.variable_scope('Branch_0'):
                max_pool_1 = max_pool_2d(input_tensor=input_tensor, pool_sizes=(3,3), stride_sizes=(2,2), paddings='VALID', names='0a_3x3')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_1 = self.conv_block(inputs=input_tensor, filters=384, kernel_size=(3,3) , strides=(2,2), padding='VALID', use_batch_norm=batch_normalization, name='0b_3x3')
            with tf.compat.v1.variable_scope('Branch_2'):
                conv_2 = self.conv_block(inputs=input_tensor, filters=256, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0c_1x1')
                conv_3 = self.conv_block(inputs=conv_2, filters=256, kernel_size=(3,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1c_3x3')
                conv_4 = self.conv_block(inputs=conv_3, filters=384, kernel_size=(3,3) , padding='VALID', strides=(2,2), use_batch_norm=batch_normalization, name='2c_1x1')
            with tf.compat.v1.variable_scope('Concatenate_1'):
                concat_1 = concatenate(list_tensor=[max_pool_1,conv_1,conv_4], axis=-1, names='maxpool_3x3_conv_3x3_conv_1x1')
        return concat_1
    
    
    # Inception B
    def inception_resnet_b(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {[type]} -- [description] (default: {None})
        """
        with tf.compat.v1.variable_scope(scope, 'BlockInceptionResnetB', [input_tensor], reuse=reuse):
            with tf.compat.v1.variable_scope('Relu_1'):
                relu_1 = relu(input_tensor=input_tensor, names='input_inception_resnet_b')
            with tf.compat.v1.variable_scope('Branch_0'):
                conv_1 = self.conv_block(inputs=relu_1, filters=192, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0a_1x1')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_2 = self.conv_block(inputs=relu_1, filters=128, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0b_1x1')
                conv_3 = self.conv_block(inputs=conv_2, filters=160, kernel_size=(1,7) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1b_1x7')
                conv_4 = self.conv_block(inputs=conv_3, filters=192, kernel_size=(7,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='2b_7x1')
            with tf.compat.v1.variable_scope('Concatenate_1'):
                concat_1 = concatenate(list_tensor=[conv_1,conv_4], axis=-1, names='maxpool_3x3_conv_3x3_conv_1x1')
            with tf.compat.v1.variable_scope('Branch_1c'):
                conv_5 = self.conv_block(inputs=concat_1, filters=1154, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0c_1x1')

            concat_1 = concatenate(list_tensor=[conv_1,conv_2,conv_5,conv_10], axis=-1, names=None)
        return concat_1

    # Reduction B
    def reduction_b(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {[type]} -- [description] (default: {None})
        """
        with tf.compat.v1.variable_scope(scope, 'BlockReductionB', [input_tensor], reuse=reuse):
            with tf.compat.v1.variable_scope('Branch_0'):
                max_pool_1 = max_pool_2d(input_tensor=input_tensor, pool_sizes=(3,3), stride_sizes=(2,2), paddings='VALID', names='0a_3x3')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_1 = self.conv_block(inputs=input_tensor, filters=192, kernel_size=(1,1) , strides=(1,1), padding='SAME', use_batch_norm=batch_normalization, name='0b_1x1')
                conv_2 = self.conv_block(inputs=conv_1, filters=192, kernel_size=(3,3) , strides=(2,2), padding='VALID', use_batch_norm=batch_normalization, name='0b_1x1')
            with tf.compat.v1.variable_scope('Branch_2'):
                conv_3 = self.conv_block(inputs=input_tensor, filters=256, kernel_size=(1,1) , strides=(1,1), padding='SAME', use_batch_norm=batch_normalization, name='0c_1x1')
                conv_4 = self.conv_block(inputs=conv_3, filters=256, kernel_size=(1,7) , strides=(1,1), padding='SAME', use_batch_norm=batch_normalization, name='1c_1x7')
                conv_5 = self.conv_block(inputs=conv_4, filters=320, kernel_size=(7,1) , strides=(1,1), padding='SAME', use_batch_norm=batch_normalization, name='2c_7x1')
                conv_6 = self.conv_block(inputs=conv_5, filters=320, kernel_size=(3,3) , strides=(2,2), padding='VALID', use_batch_norm=batch_normalization, name='3c_3x3')
            
            concat_1 = concatenate(list_tensor=[max_pool_1,conv_2,conv_6], axis=-1, names=None)
        return concat_1

    # Inception C
    def inception_c(self, input_tensor, batch_normalization=True, scope=None, reuse=True):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            batch_normalization {bool} -- [description] (default: {True})
            scope {[type]} -- [description] (default: {None})
            reuse {[type]} -- [description] (default: {None})
        """
        with tf.compat.v1.variable_scope(scope, 'BlockInceptionC', [input_tensor], reuse=reuse):
            with tf.compat.v1.variable_scope('Branch_0'):
                avg_pool_1 = avarage_pool_2d(input_tensor=input_tensor, kernel_sizes=(3,3), padding='SAME', stride_sizes=(1,1), names='0a_3x3')
                conv_1 = self.conv_block(inputs=avg_pool_1, filters=256, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1a_1x1')
            with tf.compat.v1.variable_scope('Branch_1'):
                conv_2 = self.conv_block(inputs=input_tensor, filters=256, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0b_1x1')
            with tf.compat.v1.variable_scope('Branch_2'):
                conv_3 = self.conv_block(inputs=input_tensor, filters=384, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0c_1x1')
                with tf.compat.v1.variable_scope('Branch_21'):
                    conv_4 = self.conv_block(inputs=conv_3, filters=256, kernel_size=(1,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1c1_1x3')
                with tf.compat.v1.variable_scope('Branch_22'):
                    conv_5 = self.conv_block(inputs=conv_3, filters=256, kernel_size=(3,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1c2_3x1')
            with tf.compat.v1.variable_scope('Branch_3'):
                conv_6 = self.conv_block(inputs=input_tensor, filters=384, kernel_size=(1,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='0d_1x1')
                conv_7 = self.conv_block(inputs=conv_6, filters=448, kernel_size=(1,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='1d_1x3')
                conv_8 = self.conv_block(inputs=conv_7, filters=512, kernel_size=(3,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='2d_3x1')
                with tf.compat.v1.variable_scope('Branch_31'):
                    conv_9 = self.conv_block(inputs=conv_8, filters=256, kernel_size=(3,1) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='3d1_3x1')
                with tf.compat.v1.variable_scope('Branch_32'):
                    conv_10 = self.conv_block(inputs=conv_8, filters=256, kernel_size=(1,3) , padding='SAME', strides=(1,1), use_batch_norm=batch_normalization, name='3d2_1x3')

            concat_1 = concatenate(list_tensor=[conv_1,conv_2,conv_4,conv_5,conv_9,conv_10], axis=-1, names=None)
        return concat_1

    # Create Model
    def create_model(self):
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
        
        Keyword Arguments:
            classes {int} -- [description] (default: {1000})
            n_inception_a {int} -- [description] (default: {4})
            n_inception_b {int} -- [description] (default: {7})
            n_inception_c {int} -- [description] (default: {3})
            batch_normalizations {bool} -- [description] (default: {True})
        """
        with tf.compat.v1.variable_scope(self.scope, 'inception_v4_resnet_v2', [self.input_tensor]):
            stem_layer = self.stem_block(input_tensor=self.input_tensor, batch_normalization=self.batch_normalizations)

            for i in range(0,self.n_inception_a):
                inception_a_layer = self.inception_a(input_tensor=stem_layer, batch_normalization=self.batch_normalizations)

            reduction_a_layer = self.reduction_a(input_tensor=inception_a_layer, batch_normalization=self.batch_normalizations)

            for j in range(0,self.n_inception_b):
                inception_b_layer = self.inception_b(input_tensor=reduction_a_layer, batch_normalization=self.batch_normalizations)

            reduction_b_layer = self.reduction_b(input_tensor=inception_b_layer, batch_normalization=self.batch_normalizations)

            for k in range(0,self.n_inception_c):
                inception_c_layer = self.inception_c(input_tensor=reduction_b_layer, batch_normalization=self.batch_normalizations)

            if n_inception_a == 0:
                inception_v4 = stem_layer
            elif n_inception_b == 0:
                inception_v4 = reduction_a_layer
            elif n_inception_c == 0:
                inception_v4 = reduction_b_layer
            else:
                inception_v4 = inception_c_layer

            base_var_list = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.GLOBAL_VARIABLES)

            inception_v4 = flatten(input_tensor=inception_v4, names='output')

            inception_v4 = avarage_pool_2d(input_tensor=inception_v4, kernel_sizes=(3,3), padding='SAME', stride_sizes=(1,1), names='output')

            inception_v4 =  dropout(input_tensor=inception_v4, names='output', dropout_rates=self.dropout_rate)

            non_logit = dense(input_tensor=inception_v4, hidden_units=self.classes, names='output')
            
            if self.classes > 2:
                output = softmax(input_tensor=non_logit, names='output')
            else:
                output = sigmoid(input_tensor=non_logit, names='output')

            full_var_list = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.GLOBAL_VARIABLES)
        return non_logit, output, base_var_list, full_var_list




import tensorflow as tf
from poda.layers.activation import *
from poda.layers.dense import *
from poda.layers.regularizer import *

def avarage_pool_1d(input_tensor, kernel_sizes=(3), stride_sizes=(1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
    
    Keyword Arguments:
        kernel_sizes {tuple} -- [Size of kernel] (default: {(3,3)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[type]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [A layer with avarage pool 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'avg_pool_1d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.avg_pool1d(value=input_tensor,ksize=kernel_sizes,strides=stride_sizes,padding=paddings,data_format='NHWC',name=names)
    return layer

def avarage_pool_2d(input_tensor, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
    
    Keyword Arguments:
        kernel_sizes {tuple} -- [Size of kernel] (default: {(3,3)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[type]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [A layer with avarage pool 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'avg_pool_2d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.avg_pool(value=input_tensor,ksize=kernel_sizes,strides=stride_sizes,padding=paddings,data_format='NHWC',name=names)
    return layer

def avarage_pool_3d(input_tensor, kernel_sizes=(3,3,3), stride_sizes=(1,1,1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
    
    Keyword Arguments:
        kernel_sizes {tuple} -- [Size of kernel] (default: {(3,3)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[type]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [A layer with avarage pool 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'avg_pool_3d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.avg_pool3d(value=input_tensor,ksize=kernel_sizes,strides=stride_sizes,padding=paddings,data_format='NHWC',name=names)
    return layer

def batch_normalization(input_tensor, is_trainable=True, decay = 0.999, epsilon = 1e-3):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer]
    
    Keyword Arguments:
        is_trainable {bool} -- [State of trainable layer] (default: {True})
    
    Returns:
        [Tensor] -- [A trainable tensor]
    """
    scale = tf.Variable(tf.ones([input_tensor.get_shape()[-1]]))
    beta = tf.Variable(tf.zeros([input_tensor.get_shape()[-1]]))
    pop_mean = tf.Variable(tf.zeros([input_tensor.get_shape()[-1]]), trainable=False)
    pop_var = tf.Variable(tf.ones([input_tensor.get_shape()[-1]]), trainable=False)

    if is_trainable:
        batch_mean, batch_var = tf.nn.moments(input_tensor,[0])
        train_mean = tf.assign(pop_mean, pop_mean * decay + batch_mean * (1 - decay))
        train_var = tf.assign(pop_var, pop_var * decay + batch_var * (1 - decay))

        with tf.control_dependencies([train_mean, train_var]):
            return tf.nn.batch_normalization(input_tensor, batch_mean, batch_var, beta, scale, epsilon)
    else:
        return tf.nn.batch_normalization(input_tensor, pop_mean, pop_var, beta, scale, epsilon)

def convolution_1d(input_tensor, number_filters, kernel_sizes=3, stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
        number_filters {[int]} -- [the dimensionality of the output space (i.e. the number of filters in the convolution).]
    
    Keyword Arguments:
        kernel_sizes {int} -- [Size of kernel] (default: {3})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        activation {str} -- [Type of activation function in layer] (default: {'relu'})
        dropout_layers {[type]} -- [Value of dropout rate and determine to use dropout or not] (default: {None})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [Layer convolution 1D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'conv_1d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    weight = new_weights(shapes=[kernel_sizes, input_tensor.get_shape().as_list()[-1], number_filters], names=names)

    layer = tf.nn.conv1d(value=input_tensor, filters=weight, stride=stride_sizes[0], padding=paddings, use_cudnn_on_gpu=True, data_format=None, name='conv_1d_'+names)

    if activations!=None:
        layer = define_activation_function(input_tensor=layer, activation_name=activations, names=names)
    else:
        layer = layer

    if dropout_layers!=None:
        layer = dropout(input_tensor=layer, names=names)
    else:
        layer = layer
    return layer

def convolution_2d(input_tensor, number_filters, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None):
    """[This function creates a convolution kernel that is convolved (actually cross-correlated) with the layer input to produce a tensor of outputs]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
        number_filters {[int]} -- [the dimensionality of the output space (i.e. the number of filters in the convolution).]
    
    Keyword Arguments:
        kernel_sizes {int} -- [description] (default: {(3,3)})
        stride_sizes {tuple} -- [description] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        activation {str} -- [Type of activation function in layer] (default: {'relu'})
        dropout_layer {[float]} -- [Value of dropout rate and determine to use dropout or not] (default: {None})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [Layer convolution 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'conv_2d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    weight = new_weights(shapes=[kernel_sizes[0], kernel_sizes[1], input_tensor.get_shape().as_list()[-1], number_filters], names=names)

    layer = tf.nn.conv2d(input=input_tensor, filter=weight, strides=[stride_sizes[0], stride_sizes[1]], padding=paddings, use_cudnn_on_gpu=True,
                         data_format='NHWC', dilations=[1, 1, 1, 1], name='conv_2d_'+names)

    if activations!=None:
        layer = define_activation_function(input_tensor=layer, activation_names=activations, names=names)
    else:
        layer = layer

    if dropout_layers!=None:
        layer = dropout(input_tensor=layer, names=names)
    else:
        layer = layer
    return layer

def convolution_3d(input_tensor, number_filters, kernel_sizes=(3,3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None):
    """[summary]
    
    Arguments:
        input_tensor {[type]} -- [description]
        number_filters {[type]} -- [description]
    
    Keyword Arguments:
        kernel_sizes {tuple} -- [description] (default: {(3,3,3)})
        stride_sizes {tuple} -- [description] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        activations {str} -- [Type of activation function in layer] (default: {'relu'})
        dropout_layers {[type]} -- [Value of dropout rate and determine to use dropout or not] (default: {None})
        names {[type]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [Layer convolution 1D with dtype tf.float32]
    """    
    if names!=None:
        names = str(names)
    else:
        names = 'conv_3d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    weight = new_weights(shapes=[kernel_sizes[0], kernel_sizes[1], kernel_sizes[1], input_tensor.get_shape().as_list()[-1], number_filters], names=names)

    layer = tf.nn.conv3d(input,filter,strides=[stride_sizes[0],stride_sizes[1],stride_sizes],padding=paddings,data_format='NDHWC',dilations=[1, 1, 1, 1, 1],name='conv_3d_'+names)
    
    if activations!=None:
        layer = define_activation_function(input_tensor=layer, activation_name=activations, names=names)
    else:
        layer = layer
    
    if dropout_layers!=None:
        layer = dropout(input_tensor=layer, names=names)
    else:
        layer = layer

    return layer

def depthwise_convolution_2d(input_tensor, number_filters=1, kernel_sizes=(3,3), stride_sizes=(1,1), paddings='same', activations='relu', dropout_layers=None, names=None):
    """[Function for adding depthwise convolution 2D layer]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
        number_filters {[int]} -- [the multiplier dimensionality of the output space (i.e. the number of filters in the convolution).]
    
    Keyword Arguments:
        kernel_size {int , int} -- [Size of kernel] (default: {(3,3)})
        stride_size {int , int} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        activations {str} -- [Type of activation function in layer] (default: {'relu'})
        dropout_layer {[float]} -- [Value of dropout rate and determine to use dropout or not] (default: {None})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [Layer depthwise convolution 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'deptwise_conv_2d'
    
    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    weight = new_weights(shapes=[kernel_sizes[0], kernel_sizes[1], input_tensor.get_shape().as_list()[-1], number_filters], names=names)

    layer = tf.nn.depthwise_conv2d(input=input_tensor, filter=weight, strides=[stride_sizes[0], stride_sizes[1]], padding=paddings, rate=None, name='deptwise_conv_2d_'+names, data_format=None )

    if activations!=None:
        layer = define_activation_function(input_tensor=layer, activation_name=activations, names=names)
    else:
        layer = layer

    if dropout_layers!=None:
        layer = dropout(input_tensor=layer, names=names)
    else:
        layer = layer
    return layer

def max_pool_1d(input_tensor, pool_sizes=(2), stride_sizes=(1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
    
    Keyword Arguments:
        pool_size {tuple} -- [Size of kernel] (default: {(2,2)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[str]} -- [Name of the layer] (default: {None})

    Returns:
        [Tensor] -- [A layer with maxpool 1D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'max_pool_1d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.max_pool1d(input=input_tensor, ksize=pool_sizes, strides=stride_sizes, padding=paddings,name=None)
    return layer

def max_pool_2d(input_tensor, pool_sizes=(2,2), stride_sizes=(1,1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
    
    Keyword Arguments:
        pool_size {tuple} -- [Size of kernel] (default: {(2,2)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [A layer with maxpool 2D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'max_pool_2d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.max_pool2d(input=input_tensor, ksize=pool_sizes, strides=stride_sizes, padding=paddings,name=names)
    return layer

def max_pool_3d(input_tensor, pool_sizes=(2,2,2), stride_sizes=(1,1,1), paddings='same', names=None):
    """[summary]
    
    Arguments:
        input_tensor {[float, double, int32, int64, uint8, int16, or int8]} -- [A Tensor representing prelayer values]
    
    Keyword Arguments:
        pool_size {tuple} -- [Size of kernel] (default: {(2,2)})
        stride_sizes {tuple} -- [Size of striding of kernel] (default: {(1,1)})
        paddings {str} -- [Indicating the type of padding algorithm to use] (default: {'same'})
        names {[str]} -- [Name of the layer] (default: {None})
    
    Returns:
        [Tensor] -- [A layer with maxpool 3D with dtype tf.float32]
    """
    if names!=None:
        names = str(names)
    else:
        names = 'max_pool_3d'

    if paddings=='Valid' or paddings=='valid' or paddings=='VALID':
        paddings = 'VALID'
    elif paddings=='Same' or paddings=='same' or paddings=='SAME':
        paddings = 'SAME'
    else:
        paddings = 'SAME'

    layer = tf.nn.max_pool3d(input=input_tensor, ksize=pool_sizes, strides=stride_sizes, padding=paddings,name=None)
    return layer